# Import librairies ----------------------------------------------------------------------------------------------------
import pandas as pd
import glob
import os
import numpy as np
from pathlib import Path
import math
import time
import pyCloudCompare as cc

# get license
license_path = r"C:\Program Files\Agisoft\Metashape Pro\metashape.lic"
os.environ['agisoft_LICENSE'] = str(Path(license_path))
print('License found : ', os.environ['agisoft_LICENSE'])
import Metashape

# Custom function ------------------------------------------------------------------------------------------------------
    # create function to transform quaternion into yaw pitch roll ------------------------------------------------------
def euler_from_quaternion(x, y, z, w):
    """
    Convert a quaternion into euler angles (roll, pitch, yaw)
    roll is rotation around x in radians (counterclockwise)
    pitch is rotation around y in radians (counterclockwise)
    yaw is rotation around z in radians (counterclockwise)
    """
    t0 = +2.0 * (w * x + y * z)
    t1 = +1.0 - 2.0 * (x * x + y * y)
    roll_x = math.atan2(t0, t1)*(180/math.pi)

    t2 = +2.0 * (w * y - z * x)
    t2 = +1.0 if t2 > +1.0 else t2
    t2 = -1.0 if t2 < -1.0 else t2
    pitch_y = math.asin(t2)*(180/math.pi)

    t3 = +2.0 * (w * z + x * y)
    t4 = +1.0 - 2.0 * (y * y + z * z)
    yaw_z = math.atan2(t3, t4)*(180/math.pi)

    return -yaw_z, roll_x, -pitch_y   # in degrees in fact it's alpha nu kappa (weird)
    # find file in folder 'folder', type 'type'
    # create function to read img files in folder
    # find files with specified extension in folder
def find_files(folder, types):
    return [entry.path for entry in os.scandir(folder) if (entry.is_file() and os.path.splitext(entry.name)[1].lower() in types)]
def Nadir_ImgLoc(boundX, BoundY, wp_step, altitude, yaw, pitch, roll):
    """
    Create nadir camera plan
    :param x: X Boundary of camera plan
    :param y: Y Boundary of camera plan
    :param step: distance between 2 pic
    :param z: Flight altitude
    :param yaw:
    :param pitch:
    :param roll:
    :return: df containing X,Y,Z,yaw,pitch,roll of mission plan
    """
    # Set initial pose
    x = boundX[0]
    y = BoundY[0]
    z = altitude
   #yaw = yaw * np.pi / 180
    pitch = pitch #* np.pi / 180
    roll = roll #* np.pi / 180
    # === Generate Trajectory ===

    d = 0
    tempList = []
    while y <= BoundY[1]:
        while (d == 0 and x <= boundX[1]) or (d == 1 and x >= boundX[0]):

            # Set the pose
            yaw = d * 180 * np.pi / 180
            # fill df with XYZyawpitchroll and save it as csv
            tempList.append(
                {
                    'X': x,
                    'Y': y,
                    'Z': z,
                    'yaw': yaw,
                    'pitch': pitch,
                    'roll': roll
                }
            )

            if d == 0:
                x += wp_step
            else:
                x -= wp_step

        if d == 0:
            x -= wp_step
        else:
            x += wp_step

        d = 1 - d
        y += wp_step
    df = pd.DataFrame(tempList, columns=['X', 'Y', 'Z', 'yaw', 'pitch', 'roll'])
    return df
def get_distance(p, q):
    """
    Return euclidean distance between points p and q
    assuming both to have the same number of dimensions
    """
    # sum of squared difference between coordinates
    s_sq_difference = 0
    for p_i, q_i in zip(p, q):
        s_sq_difference += (p_i - q_i) ** 2

    # take sq root of sum of squared difference
    distance = s_sq_difference ** 0.5
    return distance
def pic_in_circle(center, radius, a, b, snapshots_number, altitude):
    '''
    parametric equation of circle in space :
    :param center_x: X coordinates of the center of the bbox
    :param center_y: Y coordinates of the center of the bbox
    :param radius: distance between corners and center of the bbox
    :param a,b: two unit vectors perpendicular to the direction of the axis and each other
    :param snapshots_number: number of markers to create
    :param altitude : altitude of the circle
    :return: df containing X and Y coordinates of selected points with lenght based on number of snapshots
    Circle in space is exprimed with X,Y AND Z : Z gives the plane where circle will be
    '''
    # Convert vectors exprimed in WGS84 back to local coordinates system
    T = ref_model.crs.localframe(ref_model.transform.translation) * ref_model.transform.matrix
    a = T.inv().mulv(a)
    b = T.inv().mulv(b)
    # altitude start from 0 : needed to subtract center altitude to "altitude" vector
    centerZ = ref_model.crs.project(ref_model.transform.matrix.mulp(center))
    altitude = T.inv().mulv(Metashape.Vector([0, 0, altitude-centerZ[2]]))
    arr=[]
    # snapshots-polygon generation
    seg = 2.0 * math.pi / snapshots_number
    for i in range(snapshots_number):
        angle = i * seg
        x = center[0] + a[0]*radius*math.cos(angle) + b[0]*radius*math.sin(angle) + altitude[0]
        y = center[1] + a[1]*radius*math.cos(angle) + b[1]*radius*math.sin(angle) + altitude[1]
        z = center[2] + a[2]*radius*math.cos(angle) + b[2]*radius*math.sin(angle) + altitude[2]
        #Create array with all the x-co and y-co of the circle
        arr.append([x,y,z])
    df = pd.DataFrame(arr)
    df.columns = ['X', 'Y','Z']
    return df
def pic_in_sphere(center, radius, snapshots_numbers):
    '''
    Take picture in sphere around object
    :param center: center of sphere
    :param radius: radius of sphere
    :param snapshots_numbers: number of picture, approximatively
    :return: df containing X,Y,Z position of img
    sphere points generation using spherical Fibonacci lattice point
    '''
    list = []
    n = 2*snapshots_numbers
    center = center
    i = np.arange(0, n, dtype=float) + 0.5
    phi = np.arccos(1 - 2*i/n)
    goldenRatio = (1 + 5**0.5)/2
    theta = 2*np.pi * i / goldenRatio
    r = radius
    x = center[0] + r * np.cos(theta) * np.sin(phi)
    y = center[1] + r * np.sin(theta) * np.sin(phi)
    z = center[2] + r * np.cos(phi)
    list.append(x)
    list.append(y)
    list.append(z)
    df = pd.DataFrame(list).transpose()
    df.columns = ['X', 'Y','Z']
    return df
def cameraTrack(chunk):
    '''
    Create camera track from camera in project, ordering by name
    :param chunk: chunk with photos
    :return: cameratrack from photos
    '''
    chunk.addCameraTrack()
    track = list()
    for camera in list(chunk.cameras):
        if (camera.type == Metashape.Camera.Type.Regular) and camera.transform:
            keyframe = chunk.addCamera()
            keyframe.type = Metashape.Camera.Type.Keyframe
            keyframe.transform = camera.transform
            track.append(keyframe)

    chunk.camera_tracks[-1].keyframes = track
    chunk.camera_track = chunk.camera_tracks[-1]


# ======================================================================================================================
# 1. INTRO PART ========================================================================================================
# ======================================================================================================================

# Enter project path
project_path = r"C:\Users\Thomas\Documents\Thomas_script_test"
model_path_in = r"C:\Users\Thomas\Documents\models"
min_altitude = 5


t = time.time()
# Open metashape project containing reference model --------------------------------------------------------------------
# open metashape georeferenced project
doc = Metashape.Document()
doc.open(path=os.path.join(project_path, "georef_rec.psx"))
ref_model = doc.chunks[0]
ref_model.label = 'reference'
crs = ref_model.crs

# initialize df containing dist C2M
C2M_df = pd.DataFrame()
# Add reference model
os.chdir(model_path_in)
models = glob.glob('*.obj')
for model in models:
    ref_model.importModel(os.path.join(model_path_in, str(model)))
    ref_model.model = None
    ref_model.model.label = str(name[0])
    name = model.split('.')
    # Create folder with model name
    model_path_out = os.path.join(project_path, str(name[0]))
    if not os.path.exists(model_path_out):
        os.makedirs(model_path_out)

    # Change view mode to textured model view
    Metashape.Application().model_view.model_view_mode = Metashape.Application.ModelView.ModelViewTextured

# ======================================================================================================================
# 2.PLAN MISSION PART ==================================================================================================
# ======================================================================================================================

    # Enter varying parameters
    overlap_list = [10, 12]

    for overlap in overlap_list:
        # create new folder containing data
        overlap_path = os.path.join(model_path_out, "Overlap" + str(overlap))
        if not os.path.exists(overlap_path):
            os.makedirs(overlap_path)

        plan_mission_task = Metashape.Tasks.PlanMission()
        # mission parameters
        plan_mission_task.sensor = 0  # choose the camera that was used for input photos (DJI for x2 test)
        plan_mission_task.min_altitude = 5
        plan_mission_task.capture_distance = 6
        plan_mission_task.horizontal_zigzags = True
        plan_mission_task.min_waypoint_spacing = 1
        plan_mission_task.overlap = overlap
        plan_mission_task.attach_viewpoints = True
        plan_mission_task.safety_distance = 1
        home is missing !
        # apply task to reference model

        plan_mission_task.apply(ref_model)
        print('mission planned with ' + str(overlap) + ' overlap !')

        # save Plan Mission as .path file ----------------------------------------------------------------------------------
        ref_model.camera_track.save(os.path.join(overlap_path, "Plan_mission.path"))
        ref_model.camera_track.label = 'MissionPlan_overlap' + str(overlap)
        ref_model.camera_track = None
        # read all .path file into one big df ------------------------------------------------------------------------------
        all_files = glob.glob(os.path.join(overlap_path, "*.path"))

        df_from_each_file = (pd.read_csv(f, header=None, delimiter='\s+') for f in all_files)
        overlap_df = pd.concat(df_from_each_file, ignore_index=True)

        overlap_df.columns = ['Time_camera', 'X', 'Y', 'Z', 'qX', 'qY', 'qZ', 'qW']
        overlap_df = overlap_df[['X', 'Y', 'Z', 'qX', 'qY', 'qZ', 'qW']]
        # save it to a csv file
        #df.to_csv(os.path.join(overlap_path, 'planMission.csv'))
        # remove .path file
        for filename in os.listdir(overlap_path):
            if filename.endswith(".path"):
                # Add your "dirname" to the file path
                os.remove(os.path.join(overlap_path, filename))

        # capture images ---------------------------------------------------------------------------------------------------
        # create new folder to save output images of model
        newpath = os.path.join(overlap_path, "picture_fromPlanMission_ov" + str(overlap))
        if not os.path.exists(newpath):
            os.makedirs(newpath)

        ''' for each line of big df :
             1. read position X,Y,Z
            2. read orientation qX,qY,qZ,qW and transform them into yaw pitch roll (for metashape python api but -alpha nu kappa for GUI metashape)
            3. Create CameraT 4x4 matrix with position and orientation infos
            4. save img in newly created folder
            in the same time, create list with pic name
            '''
        image_list = []
        img_cnt = 0
        for index, row in overlap_df.iterrows():
            position = Metashape.Vector((row['X'], row['Y'],row['Z']))
            # position vector X, Y, Z in chunk.crs 'WGS 84 + EGM96 height (EPSG::9707)'
            orientation = Metashape.Vector(
                euler_from_quaternion(row['qX'], row['qY'], row['qZ'], row['qW']))  # orientation vector

            # calculate relevant transform matrix t
            position = ref_model.crs.unproject(position)  # position in ECEF
            orientation = ref_model.crs.geogcs.localframe(position).rotation().t() * Metashape.Utils.ypr2mat(
                orientation)  # orientation matrix in ECEF
            transform = Metashape.Matrix.Translation(position) * Metashape.Matrix.Rotation(orientation)
            transform = ref_model.transform.matrix.inv() * transform * Metashape.Matrix.Diag((1, -1, -1, 1))
            cameraT = Metashape.Matrix.Translation(transform.translation()) * Metashape.Matrix.Rotation(
                transform.rotation())  # 4x4 transform matrix 3 translations and 3 rotations

            # Capture img to new folder
            image = ref_model.model.renderImage(cameraT, ref_model.sensors[0].calibration)
            print("saved img " + str(index + 1) + " on " + str(len(overlap_df.index)) + " to " + str(newpath))
            image.save(os.path.join(newpath, "pic" + str(index + 1) + ".jpg"))  # so first pic = pic1
            image_list.append('pic' + str(index + 1) + '.jpg')
            img_cnt += 1

        # create df with pic name, X, Y, Z position of virtual picture to add to reference panel of reconstruction chunk
        # save it to imgLoc.csv
        overlap_df = overlap_df[['X', 'Y', 'Z']]
        overlap_df['img_name'] = image_list  # add img_list as column
        # move column to 1st place
        img_n = overlap_df['img_name']
        overlap_df.drop(labels=['img_name'], axis=1, inplace=True)
        overlap_df.insert(0, 'img_name', img_n)
        overlap_df.to_csv(os.path.join(overlap_path, 'overlap'+str(overlap)+'imgLoc.csv'), index=False)

        # Recreate 3D representation based on captured images ----------------------------------------------------------
        photos = find_files(newpath, [".jpg", ".jpeg", ".tif", ".png"])
        reconstruction = doc.addChunk()
        reconstruction.label = 'reconstruction_overlap' + str(overlap)

        reconstruction.addPhotos(photos)
        print(str(len(reconstruction.cameras)) + " images loaded")

        #Align Cameras
        reconstruction.matchPhotos(keypoint_limit=80000, tiepoint_limit=30000, generic_preselection=True,
                                       reference_preselection=True)
        reconstruction.alignCameras()
        doc.save(path=os.path.join(overlap_path, "reconstructed_modelsOverlap" + str(overlap)+".psx"))
        # Build depth maps
        reconstruction.buildDepthMaps(downscale=2, filter_mode=Metashape.MildFiltering)
        doc.save(path=os.path.join(overlap_path, "reconstructed_modelsOverlap" + str(overlap)+".psx"))
        # Build dense clouds
        reconstruction.buildDenseCloud(point_confidence=True)
        doc.save(path=os.path.join(overlap_path, "reconstructed_modelsOverlap" + str(overlap)+".psx"))

        # Add reference in the same crs
        reconstruction.importReference(path=os.path.join(overlap_path, 'overlap'+str(overlap)+'imgLoc.csv'), crs=crs,
                                       format=Metashape.ReferenceFormatCSV, columns='nxyz', delimiter=',', skip_rows=1)
        reconstruction.updateTransform()
        reconstruction.crs = crs

        # Save both model as point of cloud and mesh + save a report of reconstruction process -------------------------
        reconstruction.exportPoints(path=os.path.join(overlap_path, "reconstruction_pc_ov" + str(overlap) + ".las"),
                                                        source_data=Metashape.PointCloudData)
        if reconstruction.dense_cloud:  # if existing, export dense cloud
            reconstruction.exportPoints(path=os.path.join(overlap_path, "reconstruction_dc_ov" + str(overlap) + ".las"),
                                                            source_data=Metashape.DenseCloudData)
            reconstruction.exportPoints(path=os.path.join(overlap_path, "reconstruction_dc_ov" + str(overlap) + "CRS_WGS84.las"),
                                                            source_data=Metashape.DenseCloudData, crs=crs)
        # Make a report of the reconstruction process
        reconstruction.exportReport(path=os.path.join(overlap_path, 'report_reconstruction_ov' + str(overlap) + '.pdf'))
        doc.save(path=os.path.join(overlap_path, "reconstructed_modelsOverlap" + str(overlap)+".psx"))

        # Cloudcompare : reconstructions comparaison -------------------------------------------------------------------
        cli = cc.CloudCompareCLI()
        cmd = cli.new_command()
        # cmd.silent()  # Disable console
        cmd.open(os.path.join(overlap_path, "reconstruction_dc_ov" + str(overlap) + ".las"))
        cmd.open(os.path.join(model_path_in, str(model)))
        cmd.match_centers()
        cmd.icp(overlap=80, adjust_scale=True)
        cmd.c2m_dist(max_dist=0.1)
        cmd.cloud_export_format(cc.CLOUD_EXPORT_FORMAT.ASCII, extension="xyz", add_header=True)
        cmd.save_clouds(os.path.join(overlap_path, "C2M_pc_ov" + str(overlap) + "_maxDist10cm.xyz"))
        print(cmd)
        cmd.execute()

        cmd2 = cli.new_command()
        # cmd.silent()  # Disable console
        cmd2.open(os.path.join(overlap_path, "reconstruction_dc_ov" + str(overlap) + ".las"))
        cmd2.open(os.path.join(model_path_in, str(model)))
        cmd2.match_centers()
        cmd2.icp(overlap=80, adjust_scale=True)
        cmd2.c2m_dist()
        cmd2.cloud_export_format(cc.CLOUD_EXPORT_FORMAT.ASCII, extension="xyz", add_header=True)
        cmd2.save_clouds(os.path.join(overlap_path, "C2M_pc_ov" + str(overlap) + ".xyz"))
        print(cmd2)
        cmd2.execute()

        # C2M analysis -------------------------------------------------------------------------------------------------
        df = pd.read_csv(os.path.join(overlap_path, "C2M_pc_ov" + str(overlap) + ".xyz"), header=0, delimiter='\s+')
        dist_C2M = df.iloc[:,-1:]
        dist_C2M = dist_C2M.sample(n=1500000)
        C2M_df['C2M_dist_ov'+str(overlap)] = dist_C2M

        elapsed = time.time() - t
        # Save some info's to txt file -----------------------------------------------------------------------------------------
        file = open(os.path.join(overlap_path, "reconstruction_infos_ov" + str(overlap) + ".txt"), "w")
        file.write("Overlap of reconstruction : " + str(overlap) + "% \n")
        file.write("Elapsed time to run script = " + str(elapsed) + " seconds \n")
        file.write("number of pic taken = " + str(img_cnt) + " images ")
        file.close

# ======================================================================================================================
# 3. NADIR, CIRCLE AND SPHERE PREP PART ================================================================================
# ======================================================================================================================
    # Remove tie points
    ref_model.point_cloud = None
    # Add dense cloud from last overlap reconstruction
    ref_model.importPoints(path=os.path.join(overlap_path, "reconstruction_dc_ov" + str(overlap) + "CRS_WGS84.las"), crs=crs)
    # Reset region so that bbox in now around reference model
    ref_model.resetRegion()
    # place markers in each corners
    region = ref_model.region
    list = []
    for i in range(8):  # finding bounding box corners
        pos = Metashape.Vector(
            [region.size.x * ((i & 1) - 0.5), 0.5 * region.size.y * ((i & 2) - 1),
             0.25 * region.size.z * ((i & 4) - 2)])
        pos = region.center + region.rot * pos  # coordinates in the internal coordinates
        corner = ref_model.addMarker()
        corner.label = "corner " + str(i + 1)
        corner.reference.location = ref_model.crs.project(ref_model.transform.matrix.mulp(pos))
        corner.reference.enabled = True
        # Save corner location to list
        list.append([corner.reference.location[0], corner.reference.location[1], corner.reference.location[2]])
    # Convert lists of corners to df
    df_corners = pd.DataFrame(list, columns=['X', 'Y', 'Z'])

    # add marker in center of bbox
    center = ref_model.addMarker()
    center.label = 'center'
    center.reference.location = ref_model.crs.project(ref_model.transform.matrix.mulp(region.center))

    # Compute Radius distance btw points and center of bbox
    radius = get_distance(center.position, corner.position)

# ======================================================================================================================
# 4. NADIR PART ========================================================================================================
# ======================================================================================================================

    # Set waypoint bounds and parameters for a lawnmower pattern
    wp_bound_x = [min(df_corners['X']), max(df_corners['X'])]
    wp_bound_y = [min(df_corners['Y']), max(df_corners['Y'])]
    wp_step = (max(df_corners['Y']) - min(df_corners['Y'])) / 10
    wp_z = max(df_corners['Z'] + 5)

    pitch_list = [0,-45]
    for pitch in pitch_list:
        Nadir = Nadir_ImgLoc(wp_bound_x, wp_bound_y, wp_step, wp_z, 0, pitch, 0)

        Nadir_path = os.path.join(model_path_out, "NADIR_pitch" + str(pitch))
        if not os.path.exists(Nadir_path):
            os.makedirs(Nadir_path)

        newpath = os.path.join(Nadir_path, "picture_fromNADIR_pitch" + str(pitch))
        if not os.path.exists(newpath):
            os.makedirs(newpath)

        image_list = []
        img_cnt = 0
        for index, row in Nadir.iterrows():
            position = Metashape.Vector((row['X'], row['Y'], row[
                'Z']))  # position vector X, Y, Z in chunk.crs 'WGS 84 + EGM96 height (EPSG::9707)'
            orientation = Metashape.Vector((row['yaw'], row['pitch'], row['roll']))  # orientation vector

            # calculate relevant transform matrix t
            position = ref_model.crs.unproject(position)  # position in ECEF
            orientation = ref_model.crs.geogcs.localframe(position).rotation().t() * Metashape.Utils.ypr2mat(
                orientation)  # orientation matrix in ECEF
            transform = Metashape.Matrix.Translation(position) * Metashape.Matrix.Rotation(orientation)
            transform = ref_model.transform.matrix.inv() * transform * Metashape.Matrix.Diag((1, -1, -1, 1))
            cameraT = Metashape.Matrix.Translation(transform.translation()) * Metashape.Matrix.Rotation(
                transform.rotation())  # 4x4 transform matrix 3 translations and 3 rotations

            # Capture img to new folder
            image = ref_model.model.renderImage(cameraT, ref_model.sensors[0].calibration)
            print("saved img " + str(index + 1) + " on " + str(len(Nadir.index)) + " to " + str(newpath))
            image.save(os.path.join(newpath, "pic" + str(index + 1) + ".jpg"))  # so first pic = pic1
            image_list.append('pic' + str(index + 1) + '.jpg')
            img_cnt +=1

        Nadir['img_name'] = image_list  # add img_list as column
        # move column to 1st place
        img_n = Nadir['img_name']
        Nadir.drop(labels=['img_name'], axis=1, inplace=True)
        Nadir.insert(0, 'img_name', img_n)
        Nadir.to_csv(os.path.join(Nadir_path, 'Nadir_imgLoc.csv'), index=False)

        # Recreate 3D representation based on captured images ----------------------------------------------------------
        photos = find_files(newpath, [".jpg", ".jpeg", ".tif", ".png"])
        reconstruction = doc.addChunk()
        reconstruction.label = 'reconstruction_NADIR_pitch' + str(pitch)

        reconstruction.addPhotos(photos)
        print(str(len(reconstruction.cameras)) + " images loaded")

        # Align Cameras
        reconstruction.matchPhotos(keypoint_limit=80000, tiepoint_limit=30000, generic_preselection=True,
                                   reference_preselection=True)
        reconstruction.alignCameras()
        doc.save(path=os.path.join(Nadir_path, "reconstructed_modelsNADIR_pitch" + str(pitch) + ".psx"))
        # Build depth maps
        reconstruction.buildDepthMaps(downscale=2, filter_mode=Metashape.MildFiltering)
        doc.save(path=os.path.join(Nadir_path, "reconstructed_modelsNADIR_pitch" + str(pitch) + ".psx"))
        # Build dense clouds
        reconstruction.buildDenseCloud(point_confidence=True)
        doc.save(path=os.path.join(Nadir_path, "reconstructed_modelsNADIR_pitch" + str(pitch) + ".psx"))

        # Add reference in the same crs
        reconstruction.importReference(path=os.path.join(Nadir_path, 'Nadir_imgLoc.csv'), crs=crs,
                                       format=Metashape.ReferenceFormatCSV, columns='nxyz', delimiter=',', skip_rows=1)
        reconstruction.updateTransform()
        reconstruction.crs = crs

        # Save both model as point of cloud and mesh + save a report of reconstruction process -------------------------
        reconstruction.exportPoints(path=os.path.join(Nadir_path, "reconstruction_pc_NADIR_pitch" + str(pitch) + ".las"),
                                    source_data=Metashape.PointCloudData)
        if reconstruction.dense_cloud:  # if existing, export dense cloud
            reconstruction.exportPoints(path=os.path.join(Nadir_path, "reconstruction_dc_NADIR_pitch" + str(pitch) + ".las"),
                                        source_data=Metashape.DenseCloudData)
        # Make a report of the reconstruction process
        reconstruction.exportReport(path=os.path.join(Nadir_path, 'report_reconstruction_NADIR_pitch' + str(pitch) + '.pdf'))

        # Create CameraTrack based on img position, ordering by name
        cameraTrack(reconstruction)

        doc.save(path=os.path.join(Nadir_path, "reconstructed_modelsNADIR_pitch" + str(pitch) + ".psx"))

        # Cloudcompare : reconstructions comparison -------------------------------------------------------------------
        cli = cc.CloudCompareCLI()
        cmd = cli.new_command()
        # cmd.silent()  # Disable console
        cmd.open(os.path.join(Nadir_path, "reconstruction_dc_NADIR_pitch" + str(pitch) + ".las"))
        cmd.open(os.path.join(model_path_in, str(model)))
        cmd.match_centers()
        cmd.icp(overlap=80, adjust_scale=True)
        cmd.c2m_dist(max_dist=0.1)
        cmd.cloud_export_format(cc.CLOUD_EXPORT_FORMAT.ASCII, extension="xyz", add_header=True)
        cmd.save_clouds(os.path.join(Nadir_path, "C2M_pc_NADIR_pitch" + str(pitch) +  "_maxDist10cm.xyz"))
        print(cmd)
        cmd.execute()

        cmd2 = cli.new_command()
        # cmd.silent()  # Disable console
        cmd2.open(os.path.join(overlap_path, "reconstruction_dc_ov" + str(overlap) + ".las"))
        cmd2.open(os.path.join(model_path_in, str(model)))
        cmd2.match_centers()
        cmd2.icp(overlap=80, adjust_scale=True)
        cmd2.c2m_dist()
        cmd2.cloud_export_format(cc.CLOUD_EXPORT_FORMAT.ASCII, extension="xyz", add_header=True)
        cmd2.save_clouds(os.path.join(overlap_path, "C2M_pc_NADIR_pitch" + str(pitch) + ".xyz"))
        print(cmd2)
        cmd2.execute()

        # C2M analysis -------------------------------------------------------------------------------------------------
        df = pd.read_csv(os.path.join(Nadir_path, "C2M_pc_NADIR_pitch" + str(pitch) + ".xyz"), header=0, delimiter='\s+')
        dist_C2M = df.iloc[:,-1:]
        dist_C2M = dist_C2M.sample(n=1500000)
        C2M_df['C2M_dist_NADIR_pitch'+str(pitch)] = dist_C2M

        elapsed = time.time() - t
        # Save some info's to txt file ---------------------------------------------------------------------------------
        file = open(os.path.join(Nadir_path, "reconstruction_infos_NADIR_pitch" + str(pitch) + ".txt"), "w")
        file.write("NADIR : pitch of reconstruction : " + str(pitch) + "° \n")
        file.write("Elapsed time to run script = " + str(elapsed) + " seconds \n")
        file.write("number of pic taken = " + str(img_cnt) + " images ")
        file.close

# ======================================================================================================================
# 4. CIRCLE PART =======================================================================================================
# ======================================================================================================================
    a = Metashape.Vector([1, 0, 0])
    b = Metashape.Vector([0, 1, 0])
    altitude = [min_altitude, min_altitude + (center.reference.location[2] - min_altitude)/ 2,
                center.reference.location[2],
                center.reference.location[2] + (max(df_corners['Z']) - center.reference.location[2]) / 2,
                max(df_corners['Z'])]

    Circle_path = os.path.join(model_path_out, "Circle")
    if not os.path.exists(Circle_path):
        os.makedirs(Circle_path)

    newpath = os.path.join(Circle_path, "picture_fromPlanMission_Circle")
    if not os.path.exists(newpath):
        os.makedirs(newpath)

    # capture img in circle looking at center
    for i in altitude:
        circle = pic_in_circle(region.center, radius, a, b, 36, i)
        list = []
        image_list = []
        img_cnt = 0
        for index, row in circle.iterrows():
            location = ref_model.addMarker()
            location.reference.location = ref_model.crs.project(
                ref_model.transform.matrix.mulp(Metashape.Vector((row['X'], row['Y'], row['Z']))))
            list.append(
                [location.reference.location[0], location.reference.location[1], location.reference.location[2]])
            location = Metashape.Vector((row['X'], row['Y'], row['Z']))
            T = ref_model.crs.localframe(ref_model.transform.translation) * ref_model.transform.matrix
            direction = (ref_model.region.center - location).normalized()
            vertical = T.inv().mulv(Metashape.Vector([0, 0, 1]))
            horizontal = Metashape.Vector.cross(direction, vertical).normalized()
            vertical = Metashape.Vector.cross(direction, horizontal).normalized()
            R = Metashape.Matrix([horizontal, vertical, direction])
            cameraT = Metashape.Matrix().Translation(location) * Metashape.Matrix().Rotation(R.t())
            image = ref_model.model.renderImage(cameraT, ref_model.sensors[0].calibration)
            print("saved img " + str(index + 1) + "_alt" + str(round(i)) + " on " + str(len(circle.index)) + " to " + str(
                newpath))
            image.save(os.path.join(newpath, "pic" + str(index + 1) + "_alt" + str(round(i)) + ".jpg"))
            image_list.append("pic" + str(index + 1) + "_alt" + str(round(i)) + ".jpg")
            img_cnt +=1

        df = pd.DataFrame(list, columns=['X', 'Y', 'Z'])
        df['img_name'] = image_list  # add img_list as column
        # move column to 1st place
        img_n = df['img_name']
        df.drop(labels=['img_name'], axis=1, inplace=True)
        df.insert(0, 'img_name', img_n)
        df.to_csv(os.path.join(Circle_path, 'Circle_Alt' + str(round(i)) + 'imgLoc.csv'), index=False)

    img_cnt = img_cnt*len(altitude)

    # Recreate 3D representation based on captured images ----------------------------------------------------------
    photos = find_files(newpath, [".jpg", ".jpeg", ".tif", ".png"])
    reconstruction = doc.addChunk()
    reconstruction.label = 'reconstruction_Circle'

    reconstruction.addPhotos(photos)
    print(str(len(reconstruction.cameras)) + " images loaded")

    # Align Cameras
    reconstruction.matchPhotos(keypoint_limit=80000, tiepoint_limit=30000, generic_preselection=True,
                               reference_preselection=True)
    reconstruction.alignCameras()
    doc.save(path=os.path.join(Circle_path, "reconstructed_models_Circle.psx"))
    # Build depth maps
    reconstruction.buildDepthMaps(downscale=2, filter_mode=Metashape.MildFiltering)
    doc.save(path=os.path.join(Circle_path, "reconstructed_models_Circle.psx"))
    # Build dense clouds
    reconstruction.buildDenseCloud(point_confidence=True)
    doc.save(path=os.path.join(Circle_path, "reconstructed_models_Circle.psx"))

    # Add reference in the same crs
    reconstruction.importReference(path=os.path.join(Circle_path, 'Circle_Alt' + str(round(i)) + 'imgLoc.csv'), crs=crs,
                                   format=Metashape.ReferenceFormatCSV, columns='nxyz', delimiter=',', skip_rows=1)
    reconstruction.updateTransform()
    reconstruction.crs = crs

    # Save both model as point of cloud and mesh + save a report of reconstruction process -------------------------
    reconstruction.exportPoints(path=os.path.join(Circle_path, "reconstruction_pc_Circle.las"),
                                source_data=Metashape.PointCloudData)
    if reconstruction.dense_cloud:  # if existing, export dense cloud
        reconstruction.exportPoints(path=os.path.join(Circle_path, "reconstruction_dc_Circle.las"),
                                    source_data=Metashape.DenseCloudData)
    # Make a report of the reconstruction process
    reconstruction.exportReport(path=os.path.join(Circle_path, 'report_reconstruction_Circle.pdf'))
    # Create CameraTrack based on img position, ordering by name
    cameraTrack(reconstruction)

    doc.save(path=os.path.join(Circle_path, "reconstructed_models_Circle.psx"))

    # Cloudcompare : reconstructions comparaison -------------------------------------------------------------------
    cli = cc.CloudCompareCLI()
    cmd = cli.new_command()
    # cmd.silent()  # Disable console
    cmd.open(os.path.join(Circle_path, "reconstruction_dc_Circle.las"))
    cmd.open(os.path.join(model_path_in, str(model)))
    cmd.match_centers()
    cmd.icp(overlap=80, adjust_scale=True)
    cmd.c2m_dist(max_dist=0.1)
    cmd.cloud_export_format(cc.CLOUD_EXPORT_FORMAT.ASCII, extension="xyz", add_header=True)
    cmd.save_clouds(os.path.join(Circle_path, "C2M_pc_Circle_maxDist10cm.xyz"))
    print(cmd)
    cmd.execute()

    cmd2 = cli.new_command()
    # cmd.silent()  # Disable console
    cmd2.open(os.path.join(overlap_path, "reconstruction_dc_ov" + str(overlap) + ".las"))
    cmd2.open(os.path.join(model_path_in, str(model)))
    cmd2.match_centers()
    cmd2.icp(overlap=80, adjust_scale=True)
    cmd2.c2m_dist()
    cmd2.cloud_export_format(cc.CLOUD_EXPORT_FORMAT.ASCII, extension="xyz", add_header=True)
    cmd2.save_clouds(os.path.join(overlap_path, "C2M_pc_Circle.xyz"))
    print(cmd2)
    cmd2.execute()

    # C2M analysis -------------------------------------------------------------------------------------------------
    df = pd.read_csv(os.path.join(Circle_path, "C2M_pc_Circle.xyz"), header=0, delimiter='\s+')
    dist_C2M = df.iloc[:, -1:]
    dist_C2M = dist_C2M.sample(n=1500000)
    C2M_df['C2M_dist_Circle'] = dist_C2M

    elapsed = time.time() - t
    # Save some info's to txt file -----------------------------------------------------------------------------------------
    file = open(os.path.join(Circle_path, "reconstruction_infos_Circle.txt"), "w")
    file.write("Circle reconstruction \n")
    file.write("Elapsed time to run script = " + str(elapsed) + " seconds \n")
    file.write("number of pic taken = " + str(img_cnt) + " images ")
    file.close

# ======================================================================================================================
# 4. SPHERE PART ========================================================================================================
# ======================================================================================================================
    Sphere_path = os.path.join(model_path_out, "Sphere")
    if not os.path.exists(Sphere_path):
        os.makedirs(Sphere_path)

    newpath = os.path.join(Sphere_path, "picture_fromPlanMission_Sphere")
    if not os.path.exists(newpath):
        os.makedirs(newpath)

    sphere = pic_in_sphere(region.center, radius, 100)

    list = []
    image_list = []
    img_cnt = 0
    for index, row in sphere.iterrows():
        # Only take img of points with altitude > min_altitude
        if ref_model.crs.project(ref_model.transform.matrix.mulp(Metashape.Vector((row['X'], row['Y'], row['Z']))))[2] \
                > min_altitude:
            location = ref_model.addMarker()
            location.label = "sphere " + str(index + 1)
            location.reference.location = ref_model.crs.project(
                ref_model.transform.matrix.mulp(Metashape.Vector((row['X'], row['Y'], row['Z']))))
            list.append(
                [location.reference.location[0], location.reference.location[1], location.reference.location[2]])
            location = Metashape.Vector((row['X'], row['Y'], row['Z']))
            T = ref_model.crs.localframe(ref_model.transform.translation) * ref_model.transform.matrix
            direction = (ref_model.region.center - location).normalized()
            vertical = T.inv().mulv(Metashape.Vector([0, 0, 1]))
            horizontal = Metashape.Vector.cross(direction, vertical).normalized()
            vertical = Metashape.Vector.cross(direction, horizontal).normalized()
            R = Metashape.Matrix([horizontal, vertical, direction])
            cameraT = Metashape.Matrix().Translation(location) * Metashape.Matrix().Rotation(R.t())
            image = ref_model.model.renderImage(cameraT, ref_model.sensors[0].calibration)
            print("saved img " + str(index + 1) + " on " + str(len(sphere.index)) + " to " + str(newpath))
            print('write only half of ' + str(len(sphere.index))+', numbers are not correct')
            image.save(os.path.join(newpath, "pic" + str(index + 1) + ".jpg"))
            image_list.append("pic" + str(index + 1) + ".jpg")
            img_cnt +=1
    # save camera location to csv
    df = pd.DataFrame(list, columns=['X', 'Y', 'Z'])
    df['img_name'] = image_list  # add img_list as column
    # move column to 1st place
    img_n = df['img_name']
    df.drop(labels=['img_name'], axis=1, inplace=True)
    df.insert(0, 'img_name', img_n)
    df.to_csv(os.path.join(Sphere_path, 'Sphere_ImgLoc.csv'), index=False)

    # Recreate 3D representation based on captured images ----------------------------------------------------------
    photos = find_files(newpath, [".jpg", ".jpeg", ".tif", ".png"])
    reconstruction = doc.addChunk()
    reconstruction.label = 'reconstruction_Sphere'

    reconstruction.addPhotos(photos)
    print(str(len(reconstruction.cameras)) + " images loaded")

    # Align Cameras
    reconstruction.matchPhotos(keypoint_limit=80000, tiepoint_limit=30000, generic_preselection=True,
                               reference_preselection=True)
    reconstruction.alignCameras()
    doc.save(path=os.path.join(Sphere_path, "reconstructed_models_Sphere.psx"))
    # Build depth maps
    reconstruction.buildDepthMaps(downscale=2, filter_mode=Metashape.MildFiltering)
    doc.save(path=os.path.join(Sphere_path, "reconstructed_models_Sphere.psx"))
    # Build dense clouds
    reconstruction.buildDenseCloud(point_confidence=True)
    doc.save(path=os.path.join(Sphere_path, "reconstructed_models_Sphere.psx"))

    # Add reference in the same crs
    reconstruction.importReference(path=os.path.join(Sphere_path, 'Sphere_ImgLoc.csv'), crs=crs,
                                   format=Metashape.ReferenceFormatCSV, columns='nxyz', delimiter=',', skip_rows=1)
    reconstruction.updateTransform()
    reconstruction.crs = crs

    # Save both model as point of cloud and mesh + save a report of reconstruction process -------------------------
    reconstruction.exportPoints(path=os.path.join(Sphere_path, "reconstruction_pc_Sphere.las"),
                                source_data=Metashape.PointCloudData)
    if reconstruction.dense_cloud:  # if existing, export dense cloud
        reconstruction.exportPoints(path=os.path.join(Sphere_path, "reconstruction_dc_Sphere.las"),
                                    source_data=Metashape.DenseCloudData)
    # Make a report of the reconstruction process
    reconstruction.exportReport(path=os.path.join(Sphere_path, 'report_reconstruction_Sphere.pdf'))
    # Create CameraTrack based on img position, ordering by name
    cameraTrack(reconstruction)

    doc.save(path=os.path.join(Sphere_path, "reconstructed_models_Sphere.psx"))

    # Cloudcompare : reconstructions comparaison -------------------------------------------------------------------
    cli = cc.CloudCompareCLI()
    cmd = cli.new_command()
    # cmd.silent()  # Disable console
    cmd.open(os.path.join(Sphere_path, "reconstruction_dc_Sphere.las"))
    cmd.open(os.path.join(model_path_in, str(model)))
    cmd.match_centers()
    cmd.icp(overlap=80, adjust_scale=True)
    cmd.c2m_dist(max_dist=0.1)
    cmd.cloud_export_format(cc.CLOUD_EXPORT_FORMAT.ASCII, extension="xyz", add_header=True)
    cmd.save_clouds(os.path.join(Sphere_path, "C2M_pc_Sphere_maxDist10cm.xyz"))
    print(cmd)
    cmd.execute()

    cmd2 = cli.new_command()
    # cmd.silent()  # Disable console
    cmd2.open(os.path.join(overlap_path, "reconstruction_dc_ov" + str(overlap) + ".las"))
    cmd2.open(os.path.join(model_path_in, str(model)))
    cmd2.match_centers()
    cmd2.icp(overlap=80, adjust_scale=True)
    cmd2.c2m_dist()
    cmd2.cloud_export_format(cc.CLOUD_EXPORT_FORMAT.ASCII, extension="xyz", add_header=True)
    cmd2.save_clouds(os.path.join(overlap_path, "C2M_pc_Sphere.xyz"))
    print(cmd2)
    cmd2.execute()

    # C2M analysis -------------------------------------------------------------------------------------------------
    df = pd.read_csv(os.path.join(Sphere_path, "C2M_pc_Sphere.xyz"), header=0, delimiter='\s+')
    dist_C2M = df.iloc[:, -1:]
    dist_C2M = dist_C2M.sample(n=1500000)
    C2M_df['C2M_dist_Sphere'] = dist_C2M

    elapsed = time.time() - t
    # Save some info's to txt file -----------------------------------------------------------------------------------------
    file = open(os.path.join(Sphere_path, "reconstruction_infos_Sphere.txt"), "w")
    file.write("Sphere reconstruction \n")
    file.write("Elapsed time to run script = " + str(elapsed) + " seconds \n")
    file.write("number of pic taken = " + str(img_cnt) + " images ")
    file.close

    ref_model.remove(ref_model.markers)
    C2M_df.to_csv(os.path.join(model_path_out, 'C2M_distance_'+str(model)+'.csv'), index=False)

    elapsedmodel = time.time() - t
    file = open(os.path.join(model_path_out, "reconstruction_time_to_run_"+str(model)+".txt"), "w")
    file.write("Elapsed time to run script of model "+str(model)+" = " + str(elapsedmodel) + " seconds " + "\n")
    file.close

elapsedall = time.time() - t
file = open(os.path.join(project_path, "reconstruction_time_to_run_ALL.txt"), "w")
file.write("Elapsed time to run script = " + str(elapsedall) + " seconds " + "\n")
file.close
print("script ended ! Elapsed time is : "+str(elapsedall) + " seconds !")

